module Splines



export spline,bisectionSearch,splineInterpolate
export shapePreservingSpline,shapePreservingSpline_length,fitShapePreservingSpline
export interpolateShapePreservingSpline
#--------------------------------------------------------------------------------
#------------------------------  HELPER FUNCTIONS -------------------------------
#--------------------------------------------------------------------------------
 
function bisectionSearch(A::Array{Float64,1},key::Float64)
    n=length(A)
    accending=bool(A[n]>A[1])
    lb=1
    ub=n
    i=1
    while ub-lb >1
        m = lb+int((ub-lb)/2)
        if (key>A[m])==accending
            lb=m
        else
            ub=m
        end

        i>10 ? ub=lb : i+=1
    end
    return(lb)
end

#--------------------------------------------------------------------------------
#-------------------------------- CUBIC SPLINES ---------------------------------
#--------------------------------------------------------------------------------


function spline(x::Array{Float64,1}, y::Array{Float64,1}, yp1=Inf, ypn=Inf)
    n=length(x)
    nm1 = n-1
    u=zeros(n)
    y2=zeros(n)
    if yp1 ==Inf
        y2[1] = 0.0
        u[1]  = 0.0
    else
        y2[1] = -0.5
        u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1)
    end 

    for i = 2:nm1
        sig   = (x[i]-x[i-1])/(x[i+1]-x[i-1])
        p     = sig*y2[i-1]+2.0
        y2[i] = (sig-1.0)/p
        u[i]  = (y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1])
        u[i]  = (6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p
    end

    if (ypn == Inf)
        qn = 0.0
        un = 0.0
    else
        qn   = 0.5
        un   = (3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1])); 
    end
    
    y2[n]=(un-qn*u[n-1])/(qn*y2[n-1]+1.0)
    k=n-1
    while k>=1
        y2[k]=y2[k]*y2[k+1]+u[k]
        k=k-1
    end
    return(y2)
end


function splineInterpolate(x::Array{Float64,1}, y::Array{Float64,1},  y2::Array{Float64,1}, xVal::Float64)
    n=length(x)
    klo=int(1)
    khi=int(n)
    while (khi-klo > 1) 
        k=int((khi+klo)/2)
        if (x[k] > xVal) 
            khi=k
        else 
            klo=k
        end
    end
    h=x[khi]-x[klo]
    if h == 0.0
        yInterp= -Inf
    else
        a=(x[khi]-xVal)/h
        b=(xVal-x[klo])/h 
        yInterp=a*y[klo]+b*y[khi]+((a*a*a-a)*y2[klo]+(b*b*b-b)*y2[khi])*(h*h)/6.0
    end
    return(yInterp)
end


#--------------------------------------------------------------------------------
#------------------------ SHAPE PRESERVING SPLINES ------------------------------
#--------------------------------------------------------------------------------

type shapePreservingSpline
	k::Int64
	x::Array{Float64,1}
	y::Array{Float64,1}
  yp::Array{Float64,1}

end


function shapePreservingSpline(x::Array{Float64,1},y::Array{Float64,1})
    k=length(x)
    derivs=fitShapePreservingSpline(x,y,zeros(k),true)
    yp=derivs[:,1]
    shapePreservingSpline(k,x,y,yp)
  end


function shapePreservingSpline_length(s::shapePreservingSpline)
  return(s.k)
end

#------------------------

function fitShapePreservingSpline(x::Array{Float64,1},y::Array{Float64,1},yprime::Array{Float64,1},Schumaker=true)


  # Returns an array of coeffients
  # 1-D piecewise monotone quadratic spline interpolation
  # Adapted from code posted by Joe Henning - Fall 2012

  #   On Shape-Preserving Quadratic Spline Interpolation
  #   Schumaker, Larry L.
  #   SIAM Journal of Numerical Analysis 20
  #   No. 4, August 1983, 854-864

  #   Monotone and Convex Quadratic Spline Interpolation
  #   Maria H. Lam
  #   Virginia Journal of Science
  #   Volume 41, Number 1, Spring 1990

  n = length(x)
  yp=yprime

  # compute slopes if necessary
  if sum(yprime)==0
    yprime = zeros(n)

    if (Schumaker)      # Schumaker (1983)
      d = zeros(n-1)
      for i   = 1:(n-1)
         h    = x[i+1] - x[i]
         dy   = y[i+1] - y[i]
         d[i] = dy/h
      end

      for i = 2:(n-1)
         if d[i-1]*d[i] > 0
            yp[i] = 2*d[i-1]*d[i]/(d[i-1] + d[i])
         else
            yp[i] = 0
         end
      end

      if d[1]*(2*d[1]-yp[2]) > 0
         yp[1] = 2*d[1] - yp[2]
      else
         yp[1] = 0
      end

      if d[n-1]*(2*d[n-1]-yp[n-1]) > 0
         yp[n] = 2*d[n-1] - yp[n-1]
      else
         yp[n] = 0
      end
    else     # Lam (1990)
      L = zeros(n-1)
      d = zeros(n-1)
      for i = 1:(n-1)
         h = x[i+1] - x[i]
         dy = y[i+1] - y[i]
         L[i] = sqrt(h*h + dy*dy)
         d[i] = dy/h
      end

      for i = 2:(n-1)
         if d[i-1]*d[i] > 0
            yp[i] = (L[i-1]*d[i-1] + L[i]*d[i])/(L[i-1] + L[i])
         else
            yp[i] = 0
         end
      end

      yp[1] = -(3*d[1] - yp[2])/2
      yp[n]= (3*d[n-1] - yp[n-1])/2
    end
  end 

  derivs=[yp]
  return(derivs)
end


#------------------------

function interpolateShapePreservingSpline(xi::Array{Float64,1},s::shapePreservingSpline)
 
  n  =length(xi)
  x  =s.x
  y  =s.y
  yp =s.yp
  yi =zeros(n)

  for i = 1:length(xi)
      klo = 1
      khi = s.k
      while khi-klo > 1
        k = int((khi+klo)/2.0) ####
        if (x[k] > xi[i])
           khi = k
        else
           klo = k
        end
      end

      h = x[khi] - x[klo]
      if (h == 0.0)
        yi[i]= 0/0
      end

      a = yp[klo]
      b = yp[khi]

      del = (y[khi]-y[klo])/h

      if (a+b)/2 == del
        eps = x[khi]
      else
        if (a-del)*(b-del) >= 0
           eps = 0.5*(x[klo]+x[khi])
        elseif abs(b-del) < abs(a-del)
           eps = x[klo] + 2*h*(b-del)/(b-a)
        else   
           eps = x[khi] + 2*h*(a-del)/(b-a)
        end
      end

      alpha = eps - x[klo]
      beta = x[khi] - eps
      shat = (2*(y[khi]-y[klo]) - alpha*a - beta*b)/h

      A1 = y[klo];
      B1 = a;
      C1 = (shat-a)/(2*alpha)
      if xi[i] <= eps
        yi[i] = A1 + B1*(xi[i]-x[klo]) + C1*(xi[i]-x[klo])*(xi[i]-x[klo])
      else
        A2 = A1 + alpha*B1 + alpha*alpha*C1
        B2 = shat
        C2 = (b-shat)/(2*beta)
        yi[i] = A2 + B2*(xi[i]-eps) + C2*(xi[i]-eps)*(xi[i]-eps)
      end
    end
    return(yi)
end

#------------------------


function interpolateShapePreservingSpline(xi::Float64,s::shapePreservingSpline)
  xii=[xi]
  yii=interpolateShapePreservingSpline(xii,s)
  return(yii[1])
end

#------------------------

function interpolateShapePreservingSpline(xi::Array{Float64,1}, x::Array{Float64,1},y::Array{Float64,1},Schumaker=true)

  n = length(x)

  # compute slopes 
  yp = zeros(n)

  if (Schumaker)      # Schumaker (1983)
    d = zeros(n-1)
    for i   = 1:(n-1)
       h    = x[i+1] - x[i]
       dy   = y[i+1] - y[i]
       d[i] = dy/h
    end

    for i = 2:(n-1)
       if d[i-1]*d[i] > 0
          yp[i] = 2*d[i-1]*d[i]/(d[i-1] + d[i])
       else
          yp[i] = 0
       end
    end

    if d[1]*(2*d[1]-yp[2]) > 0
       yp[1] = 2*d[1] - yp[2]
    else
       yp[1] = 0
    end

    if d[n-1]*(2*d[n-1]-yp[n-1]) > 0
       yp[n] = 2*d[n-1] - yp[n-1]
    else
       yp[n] = 0
    end
  else     # Lam (1990)
    L = zeros(n-1)
    d = zeros(n-1)
    for i = 1:(n-1)
       h = x[i+1] - x[i]
       dy = y[i+1] - y[i]
       L[i] = sqrt(h*h + dy*dy)
       d[i] = dy/h
    end

    for i = 2:(n-1)
       if d[i-1]*d[i] > 0
          yp[i] = (L[i-1]*d[i-1] + L[i]*d[i])/(L[i-1] + L[i])
       else
          yp[i] = 0
       end
    end

    yp[1] = -(3*d[1] - yp[2])/2
    yp[n]= (3*d[n-1] - yp[n-1])/2
  end

  yi=zeros(length(xi))

  for i = 1:length(xi)
    klo = 1
    khi = n
    while khi-klo > 1
      k = int((khi+klo)/2.0) ####
      if (x[k] > xi[i])
        khi = k
      else
        klo = k
      end
    end

    h = x[khi] - x[klo]
    if (h == 0.0)
      yi[i]= 0/0
    end

    a = yp[klo]
    b = yp[khi]

    del = (y[khi]-y[klo])/h

    if (a+b)/2 == del
      eps = x[khi]
    else
      if (a-del)*(b-del) >= 0
        eps = 0.5*(x[klo]+x[khi])
      elseif abs(b-del) < abs(a-del)
        eps = x[klo] + 2*h*(b-del)/(b-a)
      else   
        eps = x[khi] + 2*h*(a-del)/(b-a)
      end
    end

    alpha = eps - x[klo]
    beta = x[khi] - eps
    shat = (2*(y[khi]-y[klo]) - alpha*a - beta*b)/h

    A1 = y[klo]
    B1 = a
    C1 = (shat-a)/(2*alpha)
    if xi[i] <= eps
      yi[i] = A1 + B1*(xi[i]-x[klo]) + C1*(xi[i]-x[klo])*(xi[i]-x[klo])
    else
      A2 = A1 + alpha*B1 + alpha*alpha*C1
      B2 = shat
      C2 = (b-shat)/(2*beta)
      yi[i] = A2 + B2*(xi[i]-eps) + C2*(xi[i]-eps)*(xi[i]-eps)
    end
  end
  return(yi)
end

#------------------------

function interpolateShapePreservingSpline(xi::Array{Float64,1},x::Array{Float64,1}, y::Float64,Schumaker=true)
  xii=[xi]
  yii=interpolateShapePreservingSpline(xii,x,y)
  return(yii[1])
end

end