#Input:
#@results is of type struct and is a simulation results summary object
#which is the resultant  output of the simulate_CFs_fn function 
#@plot is of type boolean and if true displays a graphical output of some
#summary statistics
module SummarizeResults

#try
#    using  PlotDiagnostics ##Uses Winston
#catch
    using PlotDiagnosticsPyPlot ##Backup uses Pyplot if Winston Errors
#end
using ArrayViews
using Distributions
using Roots

using UtilityFunctions: fastIRR,calcCashFlows,mquantile,mirr
export summarize_results_fn

function summarize_results_fn(results,plt=true,plotfile="plots/result_plot.png")

    LIQHORIZON    = results.params.simu.TIMESTEPS+int(ceil(maximum(results.params.assets[1].sale_time)))
    NGUARPERS  = size(results.guarantee_draw,2)
    NYEARS     = LIQHORIZON/results.params.simu.STEPS_YR
    NCOMPOUNDS = sum(results.params.simu.initial_compounds)
    NSIMUS     = results.params.simu.NSIMUS
	
	sale_phases       = mean(results.sale_phases,1)
    withdrawal_phases = mean(results.withdrawal_phases,1)
    sale_times        = mean(results.sale_times,1) 
    withdrawal_times  = mean(results.withdrawal_times,1)


    funds_needed = mean(results.compounds_to_fund,1)
    funds_done   = mean(results.compounds_funded,1)
    funding      = [funds_needed; funds_done]
    
    A1_gt_0      = results.A1_bals[:,results.params.bonds.amort_timing[1,2]+1].>0
    A2_gt_0      = results.A2_bals[:,results.params.bonds.amort_timing[2,2]+1].>0

    p1           = mean(A1_gt_0)
    p2           = mean(A2_gt_0)
    el1          = mean(results.A1_bals[:,results.params.bonds.amort_timing[1,2]+1])./results.params.bonds.nominal[1]
    el2          = mean(results.A2_bals[:,results.params.bonds.amort_timing[2,2]+1])./results.params.bonds.nominal[2]
    loss_stats     = [p1  p2; el1  el2]

    rating1        = lookupRating(p1,3.5)   
    rating2        = lookupRating(p2,7.5)      

    ROE_raw        = results.ROE
    ROE_annualized = ((1+ROE_raw).^(1/NYEARS))-1
    ROE_mean       = mean(ROE_raw,1)
    ROE_mean_a     = mean(ROE_annualized,1)
    ROE_std_a      = std(ROE_annualized,1)
    ROE            = [ROE_mean; ROE_mean_a]
    
    ROE_quant      = mquantile(ROE_raw,[0, 0.25, 0.5, 0.75, 1])
    ROE_quant_a    = mquantile(ROE_annualized,[0, 0.25, 0.5, 0.75, 1])
    ROE_q          = [ROE_quant; ROE_quant_a]
    
    p_EQ_wipeout   = mean(ROE_annualized.==-1,1)
    p_EQ_qloss     = mean(ROE_annualized.< -0.25,1)
	p_EQ_tloss     = mean(ROE_annualized.< -0.1,1)
    p_EQ_loss      = mean(ROE_annualized.< 0,1)
    p_EQ_pos       = mean(ROE_annualized.>=0,1)
    p_EQ_05        = mean(ROE_annualized.>0.05,1)
    p_EQ_10        = mean(ROE_annualized.>=0.1,1)
    p_EQ_15        = mean(ROE_annualized.>0.15,1)
    p_EQ_25        = mean(ROE_annualized.>0.25,1)
    
    EQ_probs       = [p_EQ_wipeout,p_EQ_qloss,p_EQ_tloss,p_EQ_loss,p_EQ_pos,p_EQ_05,p_EQ_10,p_EQ_15,p_EQ_25]

    cash_flows=calcCashFlows(results)
    
    println("computing irr")
    @time IRR_quant_a = irr_cash(cash_flows)
	if any(isnan(IRR_quant_a))
		warn("NAN found in quantile.")
	end
    IRR_quant_a=IRR_quant_a[!isnan(IRR_quant_a)]
    println("done computing...$(mean(IRR_quant_a,1))")
    results.IRR     = IRR_quant_a

    p_IRR_wipeout   = mean(IRR_quant_a.==-1,1)
    p_IRR_qloss     = mean(IRR_quant_a.< -0.25,1)
    p_IRR_tloss     = mean(IRR_quant_a.< -0.1,1)
    p_IRR_loss      = mean(IRR_quant_a.< 0,1)
    p_IRR_pos       = mean(IRR_quant_a.>=0,1)
    p_IRR_05        = mean(IRR_quant_a.>0.05,1)
    p_IRR_10        = mean(IRR_quant_a.>=0.1,1)
    p_IRR_15        = mean(IRR_quant_a.>0.15,1)
    p_IRR_25        = mean(IRR_quant_a.>0.25,1)
    p_IRR_30        = mean(IRR_quant_a.>0.30,1)
    p_IRR_40        = mean(IRR_quant_a.>0.40,1)
    
    IRR_probs       = [p_IRR_wipeout,p_IRR_qloss,p_IRR_tloss,p_IRR_loss,p_IRR_pos,p_IRR_05,p_IRR_10,p_IRR_15,p_IRR_25,p_IRR_30,p_IRR_40]
    
    estModeIRR      = calcArrayMode(IRR_quant_a,2)
    IRR_CI          = round(quantile(vec(IRR_quant_a),[0.025,0.975]),3)


    @time LIRR_quant_a = irr_cash(cash_flows[:,1:LIQHORIZON,:])
	if any(isnan(LIRR_quant_a))
		warn("NAN found in quantile.")
	end
    LIRR_quant_a=LIRR_quant_a[!isnan(LIRR_quant_a)]
    p_LIRR_wipeout   = mean(LIRR_quant_a.==-1,1)
    p_LIRR_qloss     = mean(LIRR_quant_a.< -0.25,1)
    p_LIRR_tloss     = mean(LIRR_quant_a.< -0.1,1)
    p_LIRR_loss      = mean(LIRR_quant_a.< 0,1)
    p_LIRR_pos       = mean(LIRR_quant_a.>=0,1)
    p_LIRR_05        = mean(LIRR_quant_a.>0.05,1)
    p_LIRR_10        = mean(LIRR_quant_a.>=0.1,1)
    p_LIRR_15        = mean(LIRR_quant_a.>0.15,1)
    p_LIRR_25        = mean(LIRR_quant_a.>0.25,1)
    p_LIRR_30        = mean(LIRR_quant_a.>0.30,1)
    p_LIRR_40        = mean(LIRR_quant_a.>0.40,1)
    
    LIRR_probs       = [p_LIRR_wipeout,p_LIRR_qloss,p_LIRR_tloss,p_LIRR_loss,p_LIRR_pos,p_LIRR_05,p_LIRR_10,p_LIRR_15,p_LIRR_25,p_LIRR_30,p_LIRR_40]
    
    estModeLIRR      = calcArrayMode(LIRR_quant_a,2)
    LIRR_CI          = round(quantile(vec(LIRR_quant_a),[0.025,0.975]),3)


    cost_guar      =  calcGuarantee(results.guarantee_draw,0.02/2,NGUARPERS,NSIMUS)
    GUAR_quant     = quantile(vec(cost_guar),[0, .001, .002, 0.01, 0.02, 0.05, 0.10])
    prob_guar = sum(results.drew_on_guarantee.>0)/NSIMUS
    mean_cost_guar = mean(cost_guar,1)

    mean_MAX_guar  = calcGuaranteeRiskNeutral(results.drew_on_guarantee,
        results.A1_payment,results.params.ce.ce_amt,0.015/2,NGUARPERS,NSIMUS)
	
    print("\n\nv==================================v\n")
    print("\n# Simu: $(results.params.simu.NSIMUS) LiqHorizon: $NYEARS years\n")
    print("\n--------- ASSET  ANALYSIS ----------\n\n")

    print("Initial target portfoilio contains $NCOMPOUNDS compounds:\n")
    println(results.params.simu.initial_compounds)

    print("\nMean number of compounds initially purchased:\n")
    println(results.compounds_initially_bought)

    print("\nMean number of compounds exiting in each state:\n")
	print_vector(sale_phases)
	print_vector(withdrawal_phases)
    
    print("\nMean number of compounds exiting in each period (SALE FINALIZED):\n")
    print_vector(sale_times)
	print_vector(withdrawal_times)

    print("\nMean number of compounds funded in each period:\n")
    print_vector(funds_needed)
	print_vector(funds_done)

    print("\nMean cash from coverage in each period:\n")
    print_vector(mean(results.extra_cash_for_coverage,1))
    
    print("\n--------- EQUITY ANALYSIS (LIQUIDATION DATE ROE) ----------\n\n")
    print("CASH MULTIPLIER:\t\t")
    print_vector(1+ROE_mean)
    print("RAW ROE:\t\t")
    print_vector(ROE_mean)
    print("Annualized ROE:\t")
    print_pvector(ROE_mean_a)

    print("\n")
    println(hcat(["Pr. Wipeout ", "Pr. <-25%", "Pr. <-10%","Pr. Loss","Pr. Non-Loss","Pr. >5%  ","Pr. >=10%","Pr. >15%", "Pr. >25%"],EQ_probs))

    print("\n--------- EQUITY ANALYSIS (LIQUIDATION DATE IRR) ----------\n\n")
    print_mirr(cash_flows[:,1:LIQHORIZON,:])
    print("Average IRR:\t\t\t")
    print_pvector(mean(LIRR_quant_a,1))
    print("Mode IRR:\t\t\t")
    print_pvector(estModeLIRR)
    println("IRR 95% Prediction Interval\t",LIRR_CI)


    print("\n")
    println(hcat(["Pr. Wipeout", "Pr. <-25%", "Pr. <-10%","Pr. Loss","Pr. Non-Loss","Pr. >5%  ","Pr. >=10%","Pr. >15%", "Pr. >25%","Pr. >30%","Pr. >40%"],LIRR_probs))


	print("\n--------- EQUITY ANALYSIS (IRR) ----------\n\n")
    print_mirr(cash_flows)
    print("Average IRR:\t\t\t")
    print_pvector(mean(IRR_quant_a,1))
    print("Mode IRR:\t\t\t")
    print_pvector(estModeIRR)
    println("IRR 95% Prediction Interval\t",IRR_CI)


    print("\n")
    println(hcat(["Pr. Wipeout", "Pr. <-25%", "Pr. <-10%","Pr. Loss","Pr. Non-Loss","Pr. >5%  ","Pr. >=10%","Pr. >15%", "Pr. >25%","Pr. >30%","Pr. >40%"],IRR_probs))


    print("\n---------  BOND ANALYSIS  ----------\n\n")
    @printf("           \tSenior \tJunior\n")
    @printf("Prob Def.  \t%.5f \t%.5f\n",loss_stats[1,1],loss_stats[1,2])
    @printf("'Rating  ' \t%s \t%s\n",rating1, rating2)
    @printf("Exp. Loss  \t%.5f \t%.5f\n",loss_stats[2,1],loss_stats[2,2])
	
	print("\n---------  GUARANTEE ANALYSIS  ----------\n\n")
    print("\nMean amount of guarantee drawn: $mean_cost_guar\n")
    print("\nMean upper bound of guarantee drawn: $mean_MAX_guar\n")
    print("\nProbability of guarantee drawn: $prob_guar\n")
	println("Quantile\tCost\n")
	println(hcat(["Max\t", "99.9%\t", "99.8%\t", "99%\t", "98%\t", "95%\t", "90%\t"],GUAR_quant))
	
    print("\n\n^==================================^\n\n")
    
    if plt 
        try
            plot_standard_diagnostics_fn(results,plotfile)
        catch e
            println("Error during plot diagnostics...")
        end
    end


end #function summarize_results_fn


function print_vector(vec)
	for i =1:length(vec)
		@printf("%.2f ",vec[i])
	end
	print("\n")
end

function print_pvector(vec)
    for i =1:length(vec)
        @printf("%.4f ",vec[i])
    end
    print("\n")
end

function print_mirr(cash_flows)
    NEQUITY     = size(cash_flows,3)
    resmin      = zeros(NEQUITY)
    mirr_zero   = zeros(NEQUITY)
    irr_zero    = zeros(NEQUITY)
    mirrcash   = zeros(size(cash_flows,1),NEQUITY)
    mean_cash_flows = mean(cash_flows,1)

    for i = 1:NEQUITY
        if ndims(mean_cash_flows) == 3
            cflow = squeeze(mean_cash_flows[1,:,i],[1,3])
        elseif ndims(mean_cash_flows) == 1
            cflow = mean_cash_flows
        else
            error("Unexpected size of cash flow vector")
        end
        try 
            resmin[i] = fzero(x-> mirr(cflow,0.0,x)-x,0.01,1.0)
        catch
            resmin[i] = NaN
        end
        mirr_zero[i]  = (resmin[i]+1)^2-1
        
        try
            resmin[i] = fzero(x-> irr(cflow,x)-x,resmin[i]*1.2) 
        catch 
            resmin[i] = NaN
        end
        irr_zero[i]   = (resmin[i]+1)^2-1
    end
    print("MIRR (of average) 0/x:\t\t")
    print_pvector(mirr_zero)
    print("IRR (of average cash flows):\t")
    print_pvector(irr_zero)
    print("\n")
end

function irr_cash(cash_flows::Array{Float64,3})
    NEQUITY     = size(cash_flows,3)
    NSIMUS      = size(cash_flows,1)
    mirrcash   = zeros(NSIMUS,NEQUITY)

    trouble = false #non-uniqueness
    for j = 1:NSIMUS
    for i = 1:NEQUITY
        resmin = 0
        cflow=squeeze(cash_flows[j,:,i],[1,3])
        if all(cflow .<0.1)
            mirrcash[j,i]=-1.0
        else
            try resmin = fastIRR(cflow); 
            catch e
                trouble = true
                resmin = NaN
            end
            irr_zero   = (resmin+1.0)^2-1.0
            mirrcash[j,i]=irr_zero
        end
    end
    end
    if trouble
        warn("Error calculating IRR, may be due to non-uniqueness, returning NaN, probabilities may not be accurate.")
    end
    return mirrcash
end

function calcGuarantee(guarantee_draw::Array{Float64,2},disc_rate,NGUARPERS::Int64,NSIMUS::Int64)
    disc =  repmat(transpose((1+disc_rate).^(1:NGUARPERS)),NSIMUS,1)
    cost_guar = sum(guarantee_draw[:,1:NGUARPERS]./disc,2)[:,1]
end

function calcGuaranteeRiskNeutral(drew_on_guarantee::Array,A1_payment,ce_amt,disc_rate,NGUARPERS::Int64,NSIMUS::Int64)
    #=
    guartime = zeros(NGUARPERS)
    riskneutral = zeros(NGUARPERS)
    
    for i = 1:NGUARPERS
        guartime[i]=sum(drew_on_guarantee.==i)/NSIMUS
        riskneutral[i]=cdf(Normal(),quantile(Normal(),guartime[i])+(0.078/2-0.0275/2)/(0.25*sqrt(2)))
    end
    max_GUAR_AMT = maximum(sum(A1_payment,2))
    max_GUAR_AMT = min(max_GUAR_AMT,ce_amt)
    max_GUAR_AMT = max_GUAR_AMT - cumsum(A1_payment,2)

    mean_MAX_guar = sum(riskneutral./(1+disc_rate).^(1:NGUARPERS).*max_GUAR_AMT)
    =#
end

function calcArrayMode(x,precision::Int64)
    xrnd            = round(x,precision)
    pwr10           = iceil(10^precision)
    xrange          = maximum(xrnd)-minimum(xrnd)
	if xrange == Inf
		return NaN
	end
    nbins           = iceil(xrange*pwr10)
    mxCount=0
    mn=minimum(xrnd)
    estMode=mn
    curVal=mn
    delta=1.0/pwr10
    for i = 1:nbins
        nextVal=mn+(i*delta)+delta
        c=sum(sum((x.>=curVal) & (x.<nextVal) ,1))
        if c>mxCount
            mxCount=c
            estMode=curVal
        end
        curVal=nextVal
    end
    return(estMode)
end



function lookupRating(pd,maturity)
    # 
    # Currently only accurate for 3.5 yr ratings.
    # Should take maturity parameter and then interpolate (geometrically)
    # along maturity and PD (i.e., break should be for greater than geomean)
    # using default study.  
    # These numbers are from Moody's 1920-2010 study (tables 34 and 35), 
    # using alpha-numeric ratings for Aaa-Aa3 and broad ratings afterwards
    # 
    if maturity==3.5
        if pd<((0.00016+0.00048)/2) ret="Aaa-3.5 yr" 
        elseif pd<((0.00016+0.00094)/2) ret="Aa1-3.5 yr" 
        elseif pd<((0.00088+0.00192)/2) ret="Aa2-3.5 yr" 
        elseif pd<((0.00182+0.00254)/2) ret="Aa3-3.5 yr" 
        elseif pd<((0.00414+0.00632)/2) ret="A-3.5 yr" 
        elseif pd<((0.00998+0.01501)/2) ret="Baa-3.5 yr" 
        else ret="Spec-3.5yr"
        end
    else ret= "NA"
    end
    return ret
end

end
