module UtilityFunctions

using Distributions,DistributionsExtended
using Polynomial
export lognrnd,trial_cost_fn,compound_sale_price_fn,calc_amort_pmt_due_fn

export fastIRR,calcCashFlows,multiRootIRR,mirr,mquantile,sortino


function lognrnd(mu::Number,sigma::Number)
	return exp(randn()*sigma+mu)
end

#Determine cost of funding trials (stochastically) for phase comp.phase
#Input:
#@comp.phase is a cell of strings where each is 'DSC','PRE','P1','P2','P3','NDA', 'APP', 'SLD'
#@params is a struct containing the initial set of parameters set by the function init_params_fn
#Output:
#@cost is of type double and is the returned funding cost
function trial_cost_fn(dist::UnivariateDistribution,mx::Float64)
    raw_cost=rand(dist)
    return min(raw_cost,mx)
end

#Input:
#@mu is of type double and specifies the mean of a normal distribution
#@sigma is of type double and specified the standard deviation of the normal distribution
#@rho is of type double and acts as a control parameter which if 0 returns fraws from the corresponding lognormal distribution else it returns draws from a biased log normal 
#@z1 is of type double and specifies the amount of bias if rho is non-zero
#@mx is of type double and specifies the maximum value beyong which draws from the lognorma distribution are rejected
#Output:
#@value is of type double and contains the resulting draw of the log-normal distribution
function compound_sale_price_fn(dist::UnivariateDistribution,mx::Number, rho, z1)
    if mx <= 0.0
        return 0.0
    end
	ej    = randn()
    Z     = (sqrt(rho)*z1)+(sqrt(1-rho)*ej)
	U     = cdf(Normal(),Z)
	X     = quantile(dist,U)
    value = min(X,mx)'	 	
    #X     = Z.*sigma + mu
    #value = min(exp(X),mx)'
    if value < 0
        error("Negative valuation.")
    end
	return value
end


#Input:
#@per is an integer
#@a.start, and @a.end are integers representing period of start (end) of amortization
#@orig.par and @cur.par are doubles representing par at origination and outstanding
#Output:
#@pmt is a double represting the calculated principal payment due in period per
function calc_amort_pmt_due_fn(per, a_start, a_end, orig_par, cur_par)
    pmt       = 0
    amort_per = a_end - a_start + 1

    if (per>=a_start && per<=a_end)
        if (orig_par/amort_per < cur_par)
            pmt = orig_par/amort_per
        else
            pmt = cur_par
        end
    end
    return pmt;
end



function mirr(cash_flows,cost_cap,re_rate)
	if(re_rate<0) 
		return 0
	end
	m=size(cash_flows,1)
	n=size(cash_flows,2)
	result = zeros(m)
	disc_pv = 1./(1+cost_cap).^(0:n-1)
	disc_fv = (1+re_rate).^(n-1:-1:0)
	for i = 1:m
		PV = 0
		FV = 0
		for j=1:n
			PV+=max(-cash_flows[i,j],0)*disc_pv[j]
			FV+=max(cash_flows[i,j],0)*disc_fv[j]
		end
		result[i]=(FV/PV)^(1/(n-1))-1
	end
	return result
end



function calcCashFlows(results)
    cweights   = results.weight_cash
    NEQUITY    = size(cweights,2)
    NSIMUS=results.params.simu.NSIMUS
    HORIZON    = length(results.cash[1,:])
    flows = results.payoff+results.reserve_cash
    realIRR_flag = any(flows.>0) || results.params.simu.reserve_flag  || (results.params.simu.MINPAYOFFTIME <= HORIZON)
    if realIRR_flag
        for i = 1:NEQUITY
            flows[:,end,i]+=results.cash[:,end].*cweights[:,i]
            flows[:,1,i]-=results.params.bonds.nominal[end]*ones(NSIMUS)*results.params.simu.tranche_cashweights[i]
        end
    else
        warn("Using Ideal cash flow IRR calculation.")
        for i = 1:NEQUITY
            flows[:,:,i]+= ([results.cash[:,1]-results.cash_begin_per[:,1] results.cash[:,2:end]-results.cash[:,1:end-1]]).*cweights[:,i]
            flows[:,1,i]+=(results.params.bonds.nominal[1]+results.params.bonds.nominal[2])*ones(NSIMUS)*results.params.simu.tranche_cashweights[i]
        end
    end
    return flows
end

function fastIRR(cash_flows::Array{Float64,3},tol=1E-4,maxiter=50)
    NSIMUS=size(cash_flows,1)
    NEQUITY=size(cash_flows,3)
    IRR_result = zeros(NSIMUS,NEQUITY)
    for i=1:NSIMUS
    for j=1:NEQUITY
        IRR_result[i,j]=fastIRR(squeeze(cash_flows[i,:,j],[1,3]))
    end
    end
    IRR_result=(IRR_result+1.0).^2-1.0
    return IRR_result
end

##Hand coded secant method, correction factor currently disabled (didn't seem to increase performance)
function fastIRR(cash_flows::Array{Float64,1},tol=1E-4,maxiter=50)
    if all(cash_flows.<=0) 
        return -1
    end
    n=length(cash_flows)
    A=sum(cash_flows[2:end])
    C=abs(cash_flows[1])

	if C<0 && all(cash_flows[2:end]>=0)
		r1=(A/C)^(2/(n+1))-1.0
		p=log(A/C)/log(A/NPV([0; cash_flows[2:n]],r1))
		r2= (1+r1)^p-1
		kcorr = 0
		P1=0
		P2=0
		iter = 0
		while abs(r2-r1)>tol 
			P1 = NPV(cash_flows,r1)
			P2 = NPV(cash_flows,r2)
			r3 = r2+(r1-r2)*P2/(P2-P1)#*(1-1.4*(P1/(P1-3*P2+2*C))) ##secant correction factor may be faster.
			r1=r2
			r2=r3
			iter+=1
			if iter > maxiter
				error("IRR did not converge")
			end
		end
		return r2
	else
		return multiRootIRR(cash_flows)
	end
end



function irr(cash_flows::Array{Float64,1},rate)    
    return mirr(cash_flows,rate,rate)
end

function mirr(cash_flows::Array{Float64,1},cost_cap,re_rate)
    if all(cash_flows.<=0) 
        return -1
    end
    n=length(cash_flows)
    disc_pv = 1./(1+cost_cap).^(0:n-1)
    disc_fv = (1+re_rate).^(n-1:-1:0)

    PV = 0
    FV = 0
    for j=1:n
        PV+=max(-cash_flows[j],0)*disc_pv[j]
        FV+=max(cash_flows[j],0)*disc_fv[j]
    end
    return (FV/PV)^(1/(n-1))-1
end

function NPV(cash_flows::Array{Float64,1},rate)
    n=length(cash_flows)
    disc_pv = 1./(1+rate).^(0:n-1)
    PV=cash_flows'*disc_pv
    return PV[1]
end

function mquantile(v::Array{Float64,2},p::Array{Float64,1})
    n=size(v,2)
    result=zeros(n,length(p))
    for i=1:n
        result[i,:]=quantile(v[:,i],p)
    end
    return result
end

function sortino(r,T=0.15)
	n = length(r)
	DR = sqrt(sum((T-r).^2.*(r.<T)/n))
	return (mean(r)-T)/DR
end

function multiRootIRR(cash_flows::Array{Float64,1})
   coeff = Polynomial.roots(Poly(reverse(cash_flows))); # Find roots of polynomial 
   rates = (1./coeff) - 1; # Compute corresponding rates
   # Preferred rates are real-valued and positive
   ind = find((real(rates) .> 0.0) & (abs(imag(rates)) .< 1e-6));
   nind = length(ind);
   if nind==1
      # One single positive rate
      r = real(rates[ind]);
   elseif nind > 1
      # Multiple positive rates; flag stream id and return min rate
      multrates = 1;
      r= minimum(real(rates[ind]));
   else
      # Get indices of any other real rates, if any (must be <= 0)
      ind = find(abs(imag(rates)) .< 1e-6);
      nind = length(ind);
      if nind==1
         # One non-positive rate
         r = real(rates[ind]);
      elseif (nind > 1)
         # Multiple non-positive rates; flag stream id and return max rate
         multrates = 1;
         r = maximum(real(rates[ind]));
      else
		warn("NO IRR solution, returning MIRR.")
		temp = zeros(cash_flows)
        temp[1]=sum(cash_flows[cash_flows.<0])
		temp[cash_flows.>0]=cash_flows[cash_flows.>0]
		return fastIRR(temp)
      end
   end
   return r[1]
end

end
