
using DrugParams
using ParamUtilitiesOriginal
using Distributions


#This function initializes all parameters used by the simulation code
#The function contains all default values used in the code
#Input:
#Output:
#@params is of type struct and has 3 fields simu (simulation parameters), assets (asset parameters), bonds (bond parameters)
function init_params_fn()
	PHASES=["DSC","PRE","P1","P2","P3","NDA","APP"]
    PARAMETER=["vmu","vsigma","vmx","upfront","milestone","mu","sigma","mx","futurecostest"]
    # ASSET PARAMETERS
    rho = 0.2

    trans_prob_row = StateDict(1,2,3,4,5,6,7,-1)
    trans_prob_col = trans_prob_row
	    
	value = [0,7.10423,27.64386,75.5741,321.500,701.90233,817.643]
    vstdratio = 2280.0/1870.0
    vmx = [0, 100, 250, 500, 1000, 5000, 5000]
	costs = [0, 5,5,8,43,0,0]
    cstdratio = [0, 5.2/4.8, 15.2/12.8, 23.5/22.1, 86.3/60.6, 0, 0]
    cmx = [0, 20,20,100,500,0,0]
    meanDur = [NaN, 1, 1.6561, 2.0942, 2.1497, 0.8, NaN]
	success = [NaN, .69, .84, .53, .74, .96, NaN]
   	target = 5
     
    upfront = max(value-costs,0) ## to prevent arbitrage in old model
    milestone  = [0, 1, 3.75, 10, 37.6,0,0]
    ## Calculates budget for future expenses using Expectation
    budget = calBudget(costs,success,milestone,target) 

    ##Standard distribution is a capped log-normal calibrated to first and second moments.
    Distribution = "CLN"  # for Value and Costs
    
    vmu,vsigma = calCappedLN(value,value*vstdratio,vmx,0.939)
    vmu=[1.0000,1.5182,2.8833,3.9000,5.4368,6.1253,6.2889]
    valueDist = Array(UnivariateDistribution,length(PHASES))
    for i = 1:length(PHASES)
        ##Convert from LN using first,second moments
        valueDist[i]=convertLNtoGeneralized(vmu[i],vsigma[i],vmx[i],Distribution)
    end

    costDist = Array(UnivariateDistribution,length(PHASES))
    cmu,csigma = calCappedLN(costs,cstdratio.*costs,cmx)
    for i = 1:length(PHASES)
        costDist[i]=convertLNtoGeneralized(cmu[i],csigma[i],cmx[i],Distribution)
    end

    tDistType= "Geometric"
    timeMode = "Period" ##markov,period, or years (default)
    timeDist = Array(UnivariateDistribution,length(PHASES))
    for i = 1:length(PHASES)
        if !isnan(meanDur[i])
            timeDist[i]=quickTimeDist(tDistType,meanDur[i])
        end
    end
    nPhases = length(PHASES)
    #Discrete steps per year - not tested rigorously 
    trans_prob=fill(NaN,length(PHASES),length(PHASES))
	
    sell_in_phase = [-1,target,target,target,target,target,target]
    sale_delay = 0*ones(7)
	sale_time = 1.5*ones(7)
    ratio_unchanged = 1
    equity_stake    = 0.85
	assets = Assets(rho,trans_prob,trans_prob_col,
                    upfront,  milestone,
                    valueDist, 
                    [],
                    costDist, 
                    ones(nPhases), ##No structured exits
                    StrucExit(nPhases),
                    budget,
                    sell_in_phase,
					sale_delay,
                    sale_time,
                    ratio_unchanged,
                    equity_stake,
                    timeMode,
                    timeDist,
                    success,
                    1, ##no scientific minimum success
                    0 ##no scientific correlation
                    )
   
    #CE PARAMETERS
    ce_amt	   = 0.0
    ce_repay   = false
    ce_premium = 0
    ce         =Ce(ce_amt,ce_repay,ce_premium)
    
	#BOND PARAMETERS
    capital_structure = [0.15, 0.2, 0.65] #A1, A2, EQ
#   capital_structure = [0.0,0.0,1.0]
    IC_pers=2
#	interest_coverage = [1.75, 3.5, 0]  #A1, A2, EQ
    interest_coverage = [1.75, 2.75, 0]  #A1, A2, EQ
    amort_timing  = reshape([5,9,8,12],2,2)
    
    coupon = [0.05/2, 0.08/2, 0]
    servicing_rate = 0.0025
    svc_accrual_int_rate = 0.07/2
    cash_accrual_int_rate = 0.01/2

    initial_cash= 575.0
	nominal = initial_cash*capital_structure
    bonds=Bonds(capital_structure,IC_pers,interest_coverage,amort_timing,coupon,servicing_rate,
				svc_accrual_int_rate,cash_accrual_int_rate,nominal)
	
	# SIMU PARAMETERS
    NSIMUS     	= 100000
    TIMESTEPS	=  13
    STEPS_YR    = 2
	
    ## Dimensions time of acquisition, starting phase, compound type
	initial_compounds      = zeros(TIMESTEPS,length(PHASES),1)
    initial_compounds[1,:,:] = [0, 8, 8, 0, 0, 0, 0]

 	reserve_flag        = false
    fixedcost           = 0
    MINPAYOFFTIME       = typemax(Int64)
    tranche_weights     = ones(int(sum(vec(initial_compounds))),1)
    performance_fee     = NaN
    tranche_cashweights = [1.0]
    tranche_capratio = [Inf] ##do not pay tranche more than this raw ratio over initial capital

	simu=Simu(NSIMUS,TIMESTEPS,STEPS_YR,initial_cash,initial_compounds,reserve_flag,
        tranche_weights,tranche_cashweights,tranche_capratio,fixedcost,MINPAYOFFTIME,
        performance_fee,
        [],[],##no general correlation
        TIMESTEPS
        )
    
	
    #create global parameter list
    return Params(simu,[assets], bonds, ce)

end
#COPYRIGHT 2012
# This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
